/**************************************************************************
 
  ALUGrid - a library providing a mesh manager supporting simplicial
  and hexahedral meshes, local grid adaptivity for use in parallel
  computations including dynamic load balancing.

  Copyright (C) 1998 - 2002 Bernhard Schupp
  Copyright (C) 1998 - 2002 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Robert Kloefkorn
  Copyright (C) 2005 - 2012 Andreas Dedner
  Copyright (C) 2010 - 2012 Martin Nolte

  The ALUGrid library is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The ALUGrid library is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef __HEADER__XDISPLAY
#define __HEADER__XDISPLAY

// ***************************************************
// #begin(header)
// #filename:
//   xdisplay.h
// #description:
//   Basisklasse f"ur einfache Grafikausgabe 
// #classes:
//   class Xdisplay
// #end(header)
// ***************************************************

#ifdef USE_ALUGRID_XDISPLAY 
#define USE_ALUGRID_XDISPLAY 0
#endif

#if USE_ALUGRID_XDISPLAY

#include <X11/Xlib.h>
#include <X11/Xutil.h>

class Vertex;

// ***************************************************
// #begin(class)
// #description:
//   Fenster und einfache Zeichenroutinen.
// #definition:
class Xdisplay
{
  public:
    typedef struct 
    {
      Display    *display;
      Window     window;
      GC         gc;
      XEvent     event;
      KeySym     keysym;
      XSizeHints sizehints;
      Colormap   colormap;
      int        screen;
      unsigned long foreground, background;
      double scale,xoffset,yoffset;
    } display_t;
    enum {nrof_bndcols=24};
    char bnd_color[nrof_bndcols][30];
    XColor bcol[nrof_bndcols];
    char bcol_text[nrof_bndcols][30];
  private:
    display_t disp;
  public:
    Xdisplay(int argc, char **argv,
             char *text = "\0", char *windowname = "Xdisplay: output",
             double pscale = 500.0,
             double pxoffset = 125.0, double pyoffset = 600.0);
    ~Xdisplay();
    void linedraw(Vertex *, Vertex *) const;
    void linedraw(Vertex *, Vertex *, XColor) const;
    void linedraw(Vertex *, Vertex *, char [30]) const;
};
#endif
// #end(class)
// ***************************************************
#endif
