/**************************************************************************
 
  ALUGrid - a library providing a mesh manager supporting simplicial
  and hexahedral meshes, local grid adaptivity for use in parallel
  computations including dynamic load balancing.

  Copyright (C) 1998 - 2002 Bernhard Schupp
  Copyright (C) 1998 - 2002 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Robert Kloefkorn
  Copyright (C) 2005 - 2012 Andreas Dedner
  Copyright (C) 2010 - 2012 Martin Nolte

  The ALUGrid library is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The ALUGrid library is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef ALUGRID_XDISPLAY_IMP_INCLUDED
#define ALUGRID_XDISPLAY_IMP_INCLUDED

#include "grid.h"
#include "triang.h"

      
#if USE_ALUGRID_XDISPLAY 
inline void nb_draw(Xdisplay &xd,Element *el1,Element *el2) 
{
  double p1[2],p2[2];
  double bary[3]={1./3.,1./3.,1./3.};
  el1->fromlocal(bary,p1);
  el2->fromlocal(bary,p2);
  p1[0]+=0.01;
  p1[1]+=0.025;
  Fullvertex pp1(p1,-1);
  Fullvertex pp2(p2,-1);
  xd.linedraw(&pp1,&pp2,xd.bcol[12]);
}

inline void
Vtx_btree::draw(Xdisplay &xd,Element *el) {
  /*  
  if (lnb)
    nb_draw(xd,el,(Element*)lnb);
  if (rnb)
    nb_draw(xd,el,(Element*)rnb);
  */
  if (head)
    head->draw(xd,el);
}

inline void
Vtx_btree::Node::draw(Xdisplay &xd,Element *el) {
  /*
  if (lnb)
    nb_draw(xd,el,(Element*)lnb);
  if (rnb)
    nb_draw(xd,el,(Element*)rnb);
  */
  if (prev)
    prev->draw(xd,el);
  else {
    // assert(((Triang*)lnb)->leaf());
    nb_draw(xd,el,(Element*)lnb);
  }
  if (next)
    next->draw(xd,el);
  else {
    // assert(((Triang*)rnb)->leaf());
    nb_draw(xd,el,(Element*)rnb);
  }
}
#endif

#endif
