/**************************************************************************
 
  ALUGrid - a library providing a mesh manager supporting simplicial
  and hexahedral meshes, local grid adaptivity for use in parallel
  computations including dynamic load balancing.

  Copyright (C) 1998 - 2002 Bernhard Schupp
  Copyright (C) 1998 - 2002 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Robert Kloefkorn
  Copyright (C) 2005 - 2012 Andreas Dedner
  Copyright (C) 2010 - 2012 Martin Nolte

  The ALUGrid library is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The ALUGrid library is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef __ALUGRID_2D_H_INCLUDED__
#define __ALUGRID_2D_H_INCLUDED__

#define ALUGRID_NOTEMPFILE_2D
#define ALUGRID_SURFACE_2D


#ifndef ALU2DGRID_COMPATIBILITY_LEVEL
#define ALU2DGRID_COMPATIBILITY_LEVEL 1
#endif

// include all headers 
#include "stlheaders.h"

//#define COUNT_ALUGRID_FLOPS

#ifdef COUNT_ALUGRID_FLOPS
#include "double.h"
// overload original double 
//#define double Double 
#endif

#include "indexstack.h"

#ifdef COUNT_ALUGRID_FLOPS
// overload original double 
#define double Double 
#endif

#include "projectvertex.h"

namespace ALU2DGrid
{

#ifdef COUNT_ALUGRID_FLOPS
  typedef ALUGridSpace :: Double Double ;
#endif

  using namespace std;

#include "alu2d/grid.h"
#include "alu2d/triang.h"
#include "alu2d/listwalk.h"
#include "alu2d/handle.h"
#include "alu2d/vtx_btree.h"
#include "alu2d/vmmap.h"
#include "alu2d/xdisplay_imp.hh"

}


#if (ALU2DGRID_COMPATIBILITY_LEVEL < 2)
/* For compatibility with DUNE 2.0 export the old types in ALUGridSpace */

namespace ALUGridSpace
{

  typedef ALU2DGrid :: Vertex < 2 > Vertex;
  typedef ALU2DGrid :: Thinelement < 2, 3 > Thinelement;
  typedef ALU2DGrid :: Element < 2, 3 > Element;
  typedef ALU2DGrid :: Bndel < 2, 3 > Bndel;

  typedef ALU2DGrid :: Hmesh_basic < 2, 3 > Hmesh_basic;
  typedef ALU2DGrid :: Hmesh < 2, 3 > Hmesh;

  using ALU2DGrid :: Listwalkptr;
  using ALU2DGrid :: Macro;
  using ALU2DGrid :: Refco;

  typedef ALU2DGrid :: AdaptRestrictProlong2d < 2, 3 > AdaptRestrictProlong2dType;
  typedef VertexProjection < 2 , double > ProjectVertex_t;

}
#endif // #if (ALU2DGRID_COMPATIBILITY_LEVEL < 2)

#endif // #ifndef __ALUGRID_2D_H_INCLUDED__
