/**************************************************************************
 
  ALUGrid - a library providing a mesh manager supporting simplicial
  and hexahedral meshes, local grid adaptivity for use in parallel
  computations including dynamic load balancing.

  Copyright (C) 1998 - 2002 Bernhard Schupp
  Copyright (C) 1998 - 2002 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Robert Kloefkorn
  Copyright (C) 2005 - 2012 Andreas Dedner
  Copyright (C) 2010 - 2012 Martin Nolte

  The ALUGrid library is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The ALUGrid library is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef __ALUGRID_PARALLEL_CC_INCLUDED__
#define __ALUGRID_PARALLEL_CC_INCLUDED__

#include <config.h>

#include "alugrid_parallel.h"

#ifndef NDEBUG
#warning -- Compiling ALUGrid code in debug mode!!! Use -DNDEBUG to get more optimized code!!!
#endif

#define _ANSI_HEADER 


// avoid C++ bindings of MPI (-DMPIPP_H is not common enough)
// this is the only thing all MPI implementations have in common
// to do that we pretend that we are compiling C code 
#if defined(__cplusplus) 
#define rem__cplusplus __cplusplus
#undef __cplusplus
#endif

#if defined(c_plusplus) 
#define remc_plusplus c_plusplus
#undef c_plusplus
#endif

extern "C" {
  // the message passing interface (MPI) headers for C 
  #include <mpi.h>
}

// restore defines 
#if defined(rem__cplusplus) 
#define __cplusplus rem__cplusplus
#undef rem__cplusplus
#endif

#if defined(c_plusplus) 
#define c_plusplus remc_plusplus
#undef remc_plusplus
#endif


// METIS stuff 
#include "parallel/alumetis.hh"

// ParMETIS stuff 
#include "parallel/aluparmetis.hh"

namespace ALUGridSpace {

#include "parallel/gitter_pll_sti.cc"
#include "parallel/gitter_pll_impl.cc"
#include "parallel/gitter_pll_sti.cc"
#include "parallel/gitter_pll_mgb.cc"

#include "parallel/mpAccess.cc"
#include "parallel/mpAccess_MPI.cc"  

#include "parallel/gitter_pll_ldb.cc"
#include "parallel/gitter_pll_idn.cc"
// file for duneinterface 
#include "duneinterface/gitter_dune_pll_impl.cc"

} //end namespace 

#undef _ANSI_HEADER 
#endif
