/**************************************************************************
 
  ALUGrid - a library providing a mesh manager supporting simplicial
  and hexahedral meshes, local grid adaptivity for use in parallel
  computations including dynamic load balancing.

  Copyright (C) 1998 - 2002 Bernhard Schupp
  Copyright (C) 1998 - 2002 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Robert Kloefkorn
  Copyright (C) 2005 - 2012 Andreas Dedner
  Copyright (C) 2010 - 2012 Martin Nolte

  The ALUGrid library is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The ALUGrid library is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
// (c) bernhard schupp 1997 - 1998

#include "mpAccess.h"

vector < int > MpAccessLocal :: dest () const {
  vector < int > d (_linkage.size ()) ;
  for (map < int, int, less < int > > :: const_iterator i = _linkage.begin () ; 
  	i != _linkage.end () ; i ++) d [(*i).second] = (*i).first ;
  return d ;
}

void MpAccessLocal :: printLinkage (ostream & out) const {
  out << "  MpAccessLocal :: printLinkage () " << myrank () << " -> " ;
  { for (map < int, int, less < int > > :: const_iterator i = _linkage.begin () ; 
  	i != _linkage.end () ; i ++) out << (*i).first << " " ; }
  out << endl ;
  return ;
}

int MpAccessLocal :: insertRequestSymetric (set < int, less < int > > req) {
  const int me = myrank () ;
  req.erase (me) ;
  vector < int > out ;
  {for (set < int, less < int > > :: const_iterator i = req.begin () ; i != req.end () ; i ++ )
    if (_linkage.find (*i) == _linkage.end ()) out.push_back (*i) ; }
  vector < vector < int > > in = gcollect (out) ;
  { for (vector < int > :: const_iterator i = out.begin () ; i != out.end () ; i ++ )
    if (_linkage.find (*i) == _linkage.end ()) {
      int n = _linkage.size () ;
      _linkage [*i] = n ;
    }
  }
  int cnt = 0 ;
  { for (int i = 0 ; i < psize () ; i ++ )
    if (in [i].end() != find (in [i].begin(), in [i].end(), me)) {
      assert (i != me) ;
      if (_linkage.find (i) == _linkage.end ()) {
        int n = _linkage.size () ;
        _linkage [i] = n ;
        cnt ++ ; 
      }
    }
  }
  return cnt ;
}
