/**************************************************************************
 
  ALUGrid - a library providing a mesh manager supporting simplicial
  and hexahedral meshes, local grid adaptivity for use in parallel
  computations including dynamic load balancing.

  Copyright (C) 1998 - 2002 Bernhard Schupp
  Copyright (C) 1998 - 2002 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Robert Kloefkorn
  Copyright (C) 2005 - 2012 Andreas Dedner
  Copyright (C) 2010 - 2012 Martin Nolte

  The ALUGrid library is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The ALUGrid library is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
// (c) bernhard schupp 1997 - 1998
#ifndef MPACCESS_H_INCLUDED
#define MPACCESS_H_INCLUDED

#include "serialize.h"

class MpAccessGlobal {
  public :
    class CommIF
    {
    protected:
      CommIF () {}
    public:
      virtual ~CommIF() {}
      virtual CommIF* clone () const = 0;
    };

    inline virtual ~MpAccessGlobal () ;
    virtual int psize () const = 0 ;
    virtual int myrank () const = 0 ;
    virtual int barrier () const = 0 ;
    virtual int gmax (int) const = 0 ;
    virtual int gmin (int) const = 0 ;
    virtual int gsum (int) const = 0 ;
    virtual long gmax (long) const = 0 ;
    virtual long gmin (long) const = 0 ;
    virtual long gsum (long) const = 0 ;
    virtual double gmax (double) const = 0 ;
    virtual double gmin (double) const = 0 ;
    virtual double gsum (double) const = 0 ;
    virtual void gmax (double*,int,double*) const = 0 ;
    virtual void gmin (double*,int,double*) const = 0 ;
    virtual void gsum (double*,int,double*) const = 0 ;
    virtual pair<double,double> gmax (pair<double,double>) const = 0 ;
    virtual pair<double,double> gmin (pair<double,double>) const = 0 ;
    virtual pair<double,double> gsum (pair<double,double>) const = 0 ;
    virtual vector < int > gcollect (int) const = 0 ;
    virtual vector < double > gcollect (double) const = 0 ;
    virtual vector < vector < int > > gcollect (const vector < int > &) const = 0 ;
    virtual vector < vector < double > > gcollect (const vector < double > &) const = 0 ;
    virtual vector < ObjectStream > gcollect (const ObjectStream &) const = 0 ;

    //! return address of communicator (not optimal but avoid explicit MPI types here)
    virtual const CommIF* communicator() const = 0;
} ;

class MpAccessLocal : public MpAccessGlobal {
  map < int, int, less < int > > _linkage ;
  public :
    inline virtual ~MpAccessLocal () ;
    void printLinkage (ostream &) const ;
    inline void removeLinkage () ;
    inline int nlinks () const ;
    inline int link (int) const ;
    vector < int > dest () const ;
    int insertRequestSymetric (set < int, less < int > >) ;
    virtual vector < vector < int > > exchange (const vector < vector < int > > &) const = 0 ;
    virtual vector < vector < double > > exchange (const vector < vector < double > > &) const = 0 ;
    virtual vector < vector < char > > exchange (const vector < vector < char > > &) const = 0 ;
    virtual vector < ObjectStream > exchange (const vector < ObjectStream > &) const = 0 ;
    virtual void exchange (const vector < ObjectStream > &, vector < ObjectStream > &) const = 0 ;
} ;


	//
	//    #    #    #  #          #    #    #  ######
	//    #    ##   #  #          #    ##   #  #
	//    #    # #  #  #          #    # #  #  #####
	//    #    #  # #  #          #    #  # #  #
	//    #    #   ##  #          #    #   ##  #
	//    #    #    #  ######     #    #    #  ######
	//

inline MpAccessGlobal :: ~MpAccessGlobal () {
}

inline MpAccessLocal :: ~MpAccessLocal () {
}

inline int MpAccessLocal :: link (int i) const {
  assert (_linkage.end () != _linkage.find (i)) ;
  return (* _linkage.find (i)).second ;
}

inline int MpAccessLocal :: nlinks () const {
  return _linkage.size () ;
}

inline void MpAccessLocal :: removeLinkage () 
{
  _linkage.clear();
  //_linkage.erase (_linkage.begin (), _linkage.end ()) ;
  return ;
}

#endif
