/**************************************************************************
 
  ALUGrid - a library providing a mesh manager supporting simplicial
  and hexahedral meshes, local grid adaptivity for use in parallel
  computations including dynamic load balancing.

  Copyright (C) 1998 - 2002 Bernhard Schupp
  Copyright (C) 1998 - 2002 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Robert Kloefkorn
  Copyright (C) 2005 - 2012 Andreas Dedner
  Copyright (C) 2010 - 2012 Martin Nolte

  The ALUGrid library is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The ALUGrid library is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
// (c) Robert Kloefkorn 2004 - 2010 
#ifndef ALUGRIDVERTEXPROJECTION_H_INCLUDED
#define ALUGRIDVERTEXPROJECTION_H_INCLUDED

#define ALUGRID_VERTEX_PROJECTION

namespace ALUGridSpace {

// use standard namespace 
using namespace std;

// interface class for projecting vertices for boundary adjustment 
template <int dim, class coord_t = double > 
class VertexProjection
{
protected:
  // don't allow creation of an instance  
  VertexProjection () {}
public:
  // destructor 
  virtual ~VertexProjection() {}
  // projection method 
  virtual int operator()(const coord_t (&p)[dim],
                         const int segmentIndex,
                         coord_t (&ret)[dim]) const = 0;
};

} // end namespace 
#endif
