/**************************************************************************
 
  ALUGrid - a library providing a mesh manager supporting simplicial
  and hexahedral meshes, local grid adaptivity for use in parallel
  computations including dynamic load balancing.

  Copyright (C) 1998 - 2002 Bernhard Schupp
  Copyright (C) 1998 - 2002 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Robert Kloefkorn
  Copyright (C) 2005 - 2012 Andreas Dedner
  Copyright (C) 2010 - 2012 Martin Nolte

  The ALUGrid library is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The ALUGrid library is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
// (c) bernhard schupp 1997 - 1998
#ifndef KEY_H_INCLUDED
#define KEY_H_INCLUDED

template < class A > class Key3 {
  protected :
    A _a, _b, _c ;
  public :
    inline Key3 () ;
    inline Key3 (const A &,const A &,const A &) ;
    inline Key3 (const Key3 < A > &) ;
    inline const Key3 < A > & operator = (const Key3 < A > &) ;
    inline bool operator < (const Key3 < A > &) const ;
} ;

template < class A > class Key4 {
  protected :
    A _a, _b, _c, _d ;
  public :
    inline Key4 () ;
    inline Key4 (const A &, const A &, const A &, const A &) ;
    inline Key4 (const Key4 < A > &) ;
    inline const Key4 < A > & operator = (const Key4 < A > &) ;
    inline bool operator < (const Key4 < A > &) const ;
} ;

template < class A > inline Key3 < A > :: Key3 () : _a (-1), _b (-1), _c (-1) {
  return ;
}

template < class A > inline Key3 < A > :: Key3 (const A & a, const A & b, const A & c) : _a (a), _b (b), _c (c) {
  return ;
}

template < class A > inline Key3 < A > :: Key3 (const Key3 < A > & k) : _a (k._a), _b (k._b), _c (k._c) {
  return ;
}

template < class A > inline const Key3 < A > & Key3 < A > :: operator = (const Key3 < A > & k) {
  _a = k._a ;
  _b = k._b ;
  _c = k._c ;
  return * this ;
}

template < class A > inline bool Key3 < A > :: operator < (const Key3 < A > & k) const {
  return _a < k._a ? true : (_a == k._a ? (_b < k._b ? true : (_b == k._b ? (_c < k._c ? true : false) : false)) : false) ;
}

template < class A > inline Key4 < A > :: Key4 () : _a (), _b (), _c (), _d () {
  return ;
}

template < class A > inline Key4 < A > :: Key4 (const A & a, const A & b, const A & c, const A & d) : _a(a), _b (b), _c (c), _d (d) {
  return ;
}

template < class A > inline Key4 < A > :: Key4 (const Key4 < A > & k) : _a (k._a), _b (k._b), _c(k._c), _d (k._d) {
  return ;
}

template < class A > inline const Key4 < A > & Key4 < A > :: operator = (const Key4 < A > & k) {
  _a = k._a ;
  _b = k._b ;
  _c = k._c ;
  _d = k._d ;
  return * this ;
}

template < class A > inline bool Key4 < A >  :: operator < (const Key4 < A > & k) const {
  return _a < k._a ? true : (_a == k._a ? (_b < k._b ? true : (_b == k._b ? (_c < k._c ? true :
  (_c == k._c ? (_d < k._d ? true : false) : false)) : false)) : false) ;
}

#endif
